/*  File SAPROCS.C Simulated Annealing Algorithm      */
   /*   for global minimization of nonlinear multimodal functions  */

   /* Global min value must be 0 to use this version (subtract global min  */
   /*    value from obj fn to new obj fn with min of 0) */

   /* Calling programs must supply the objective function (the_obj_fn)  */
   /*   to be minimized, and also an allocated and initialized pvector.  */

   /* Algorithm based on paper by Bohachevsky, et. al., "Generalized Simulated  */
   /* Annealing...", Technometrics, 28, No.3, 1986, p. 209 */
   /* with stepsize modification by S. Welstead */

#ifndef SAPROCS_C
#define SAPROCS_C

#include <stdlib.h>
#include <time.h>
#include <conio.h>
#include <math.h>

#include "ut.h"
#include "sa.h"

void initialize_sa_weights (d_vector weights,float wt_range,int len);

void simulated_annealing (float (*the_obj_fn)(d_vector w),
			       d_vector weights,
			       int no_of_weights,
			       float wt_range,
			       float error_threshold,
			       unsigned long *final_iter,
			       float *opt_obj_value);

/* Implementation */

const float min_stepsize = 0.05;
const max_no_improvement_limit = 1000;
unsigned long iter;
float obj_value,opt_inv;
d_vector new_weights,opt_weights,direction_vector;
float beta; /* Beta is 1/temperature  */
float new_obj_value,prob,p_test,stepsize;
int no_improvement_count,no_improvement_limit;
float sub_opt_obj_value;

void initialize_sa_weights (d_vector weights,float wt_range,int len) {
     int  i;
     randomize();
     for (i = 1;i<=len;i++)
	 weights[i] = 2*wt_range*(frand() - 0.5);
     return;
     }  /* end proc */

void initialize_sa_variables (void) {
     no_improvement_count = 0;
     no_improvement_limit = 20;
     stepsize = 1.0;
     beta = 3.0;
     return;
     }  /* end proc */

void accept_this_point (d_vector weights,float opt_obj_value,int len) {
     copy_d_vector (new_weights,weights,1,len);
     obj_value = new_obj_value;
     cprintf ("\r\nIter: %lu,  error: %8.6f,  sub opt value: %8.6f,"
      "  opt error: %8.6f",iter,obj_value,sub_opt_obj_value,opt_obj_value);
     no_improvement_count++;
     return;
     } /* proc */

void simulated_annealing (float (*the_obj_fn)(d_vector w),
			       d_vector weights,
			       int no_of_weights,
			       float wt_range,
			       float error_threshold,
			       unsigned long *final_iter,
			       float *opt_obj_value) {

     int i,j,k;
     d_vector sub_opt_weights;
     float sum,denom;
     new_weights = allocate_d_vector (1,no_of_weights);
     opt_weights = allocate_d_vector (1,no_of_weights);
     sub_opt_weights = allocate_d_vector (1,no_of_weights);
     direction_vector = allocate_d_vector (1,no_of_weights);
     iter = 0;
     initialize_sa_variables();
     /* Calling program initializes weights */
     obj_value = the_obj_fn (weights);
     *opt_obj_value = obj_value;
     sub_opt_obj_value = obj_value;
     copy_d_vector (weights,opt_weights,1,no_of_weights);
     copy_d_vector (weights,sub_opt_weights,1,no_of_weights);

     while ((*opt_obj_value > error_threshold) && (!kbhit())) {
	 iter++;
	 /* set new sa direction: */
	 sum = 0.0;
	 if (stepsize >  sub_opt_obj_value)
	    stepsize = sub_opt_obj_value; /* use this only if desired min is 0 */
	 if (stepsize < min_stepsize) stepsize = min_stepsize;
	 for (i = 1;i<=no_of_weights;i++) {
	     direction_vector[i] = 2.0 * frand() - 1;
	     sum += (direction_vector[i] * direction_vector[i]);
	     }
	 denom = sqrt (sum);
	 for (i = 1;i<=no_of_weights;i++) {
	     direction_vector[i] /= denom;
	     new_weights[i] = weights[i] +
		    stepsize * direction_vector[i];
	  }  /* i */
     new_obj_value = the_obj_fn (new_weights);
     if (new_obj_value <= obj_value) {
	accept_this_point(weights,*opt_obj_value,no_of_weights);
	if (new_obj_value < sub_opt_obj_value) {
	   sub_opt_obj_value = new_obj_value;
	   copy_d_vector (new_weights,sub_opt_weights,1,no_of_weights);
	   no_improvement_count = 0;
	   denom = sub_opt_obj_value;
	   if (denom < 1e-6) denom = 1e-6;
	   no_improvement_limit = ceil(20.0/denom);
	   if (no_improvement_limit > max_no_improvement_limit)
	      no_improvement_limit = max_no_improvement_limit;
	   }  /*  if  */
	if (sub_opt_obj_value < *opt_obj_value) {
	   *opt_obj_value = sub_opt_obj_value;
	   copy_d_vector (sub_opt_weights,opt_weights,1,no_of_weights);
	   } /*  if  */
	} /* if */
      else { /* Metropolis move */
	if (fabs(*opt_obj_value) < 1e-6) opt_inv = 1e+6;
	else  opt_inv = 1.0/(*opt_obj_value);
	p_test = exp(-beta*(new_obj_value - obj_value)*opt_inv);
	prob = frand();
	if (prob < p_test) accept_this_point(weights,*opt_obj_value,
	   no_of_weights);
	} /*  else  */
     if (no_improvement_count > no_improvement_limit) {
	cprintf ("\r\n********* Resetting weights **************");
	initialize_sa_weights (weights,wt_range,no_of_weights);
	obj_value = the_obj_fn (weights);
	sub_opt_obj_value = obj_value;
	copy_d_vector (weights,sub_opt_weights,1,no_of_weights);
	initialize_sa_variables();
	}  /*  if  */
     }  /* while */

     copy_d_vector (opt_weights,weights,1,no_of_weights);
     obj_value = *opt_obj_value;
     *final_iter = iter;

     if (!kbhit()) cprintf ("%c",BEEP);
     else getch();  /* clear keyboard */
     cprintf ("\r\nPress key to clear: ");
     getch();
     clrscr();
     return;
     }  /* procedure */

#endif
